/*
 * (C) Copyright ParaSoft Corporation 2022.  All rights reserved.
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 */

#include "InterfaceInfoIterator.h"
#include "Logger.h"
#include "Error.h"

#include <string.h>
#include <wchar.h>

class MacBasedMachineId
{
public:
    MacBasedMachineId()
    {
    }

    ~MacBasedMachineId()
    {
    }

    int getMachineId()
    {
        InterfaceInfo* bestInterfaceInfo = 0;

        InterfaceInfoIterator it;
        while (it.hasNext() && !Error::hasError()) {
            InterfaceInfo* interfaceInfo = it.next();

            logInterfaceInfo(*interfaceInfo);

            if (interfaceInfo->isLoopBack) {
                Logger::getLogger().debug("  SELECTION: Ignored loopback]\n\n");
                continue;
            }
            if (calcMachineId(*interfaceInfo) == 0) {
                Logger::getLogger().debug("  SELECTION: Ignored zero machineId\n\n");
                continue;
            }

            if (bestInterfaceInfo == 0) {
                Logger::getLogger().debug("  SELECTION: Set best interface\n\n");
                bestInterfaceInfo = interfaceInfo;

            } else if (compareInterfaces(*interfaceInfo, *bestInterfaceInfo) == BETTER) {
                Logger::getLogger().debug("  SELECTION: Updated best interface\n\n");
                delete bestInterfaceInfo;
                bestInterfaceInfo = interfaceInfo;
            } else {
                Logger::getLogger().debug("  SELECTION: skipped\n\n");
            }
        }

        if (Error::hasError()) {
            return 0;
        }
        else if (bestInterfaceInfo == 0) {
            Logger::getLogger().debug("None of the interfaces matched! Will return 0\n");
            Error::setErrorMessage("No Windows interfaces");
            return 0;
        }
        else {
            int bestMachineId = calcMachineId(*bestInterfaceInfo);
            delete bestInterfaceInfo;
            return bestMachineId;
        }
    }

    static bool isMacEqual(unsigned char* mac1, unsigned char* mac2, unsigned size)
    {
        for (unsigned i = 0; (i < size) && (i < 6); ++i) {
            if (mac1[i] != mac2[i]) {
                return false;
            }
        }
        return true;
    }

private:

    enum CompareStatus
    {
        UNKNOWN,
        EQUAL,
        BETTER,
        WORSE
    };

    CompareStatus compareInterfaces(struct InterfaceInfo& iInfo1, struct InterfaceInfo& iInfo2)
    {
        if (iInfo1.isCustom) {
            if (!iInfo2.isCustom) {
                return BETTER;
            }
        }
        else if (iInfo2.isCustom) {
            // and iInfo1.isCustom is false
            return WORSE;
        }

        if (iInfo1.hasMac) {
            if (!iInfo2.hasMac) {
                return BETTER;
            }
        }
        else if (iInfo2.hasMac) {
            // and iInfo1.hasMac is false
            return WORSE;
        }

        if (iInfo1.quality != iInfo2.quality) {
            return iInfo1.quality < iInfo2.quality ? BETTER : WORSE;
        }

        if (iInfo1.name != 0) {
            if (iInfo2.name != 0) {
                int strcmpResult = strcmp(iInfo1.name, iInfo2.name);
                if (strcmpResult < 0) {
                    return BETTER;
                }
                else if (strcmpResult > 0) {
                    return WORSE;
                }
            }
            else {
                return BETTER;
            }
        }
        else if (iInfo2.name != 0) {
            return WORSE;
        }

        if (iInfo1.wname != 0) {
            if (iInfo2.wname != 0) {
                int wcscmpResult = wcscmp(iInfo1.wname, iInfo2.wname);
                if (wcscmpResult < 0) {
                    return BETTER;
                }
                else if (wcscmpResult > 0) {
                    return WORSE;
                }
            }
            else {
                return BETTER;
            }
        }
        else if (iInfo2.wname != 0) {
            return WORSE;
        }

        if (iInfo1.lowQuality != iInfo2.lowQuality) {
            return iInfo1.lowQuality < iInfo2.lowQuality ? BETTER : WORSE;
        }

        return EQUAL;
    }

    static int calcMachineId(InterfaceInfo& interfaceInfo)
    {
        if (interfaceInfo.hasMac) {
            union U {
                unsigned char c[4];
                unsigned int  n;
            } u;

            unsigned int sum = 0;
            for (int i = 0; i < 6; i++) {
                sum += interfaceInfo.mac[i];
            }
            u.c[0] = interfaceInfo.mac[1];
            u.c[1] = interfaceInfo.mac[3];
            u.c[2] = interfaceInfo.mac[5];
            u.c[3] = (unsigned char)sum;
            return u.n;

        }
        else if (interfaceInfo.hasIp) {
            return (int)*(long*)&interfaceInfo.ip[0];

        }
        else {
            return 0;
        }
    }

    static void logInterfaceInfo(InterfaceInfo& interfaceInfo)
    {
        Logger::getLogger().debug("[INTERFACE]\n");
        if (interfaceInfo.name) {
            Logger::getLogger().debug("  NAME: %s\n", interfaceInfo.name);
        }
        if (interfaceInfo.wname) {
            Logger::getLogger().wdebug(L"  WNAME: %ls\n", interfaceInfo.wname);
        }
        if (interfaceInfo.wdescription) {
            Logger::getLogger().wdebug(L"  WDESCRIPTION: %ls\n", interfaceInfo.wdescription);
        }
        if (interfaceInfo.prefix) {
            Logger::getLogger().debug("  PREFIX: %s\n", interfaceInfo.prefix);
        }
        Logger::getLogger().debug("  QUALITY: %d\n", interfaceInfo.quality);
        Logger::getLogger().debug("  CUSTOM: %s\n", interfaceInfo.isCustom ? "true" : "false");
        if (interfaceInfo.hasMac) {
            unsigned char* mac = interfaceInfo.mac;
            Logger::getLogger().debug("  MAC: %x-%x-%x-%x-%x-%x\n", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
        }
        if (interfaceInfo.hasIp) {
            unsigned char* ip = interfaceInfo.ip;
            Logger::getLogger().debug("  IP: %d.%d.%d.%d\n", ip[0], ip[1], ip[2], ip[3]);
        }
        Logger::getLogger().debug("  MACHINEID: %x\n", calcMachineId(interfaceInfo));
    }

    friend class MacBasedMachineIdTest;
};
