/*
 * (C) Copyright ParaSoft Corporation 2022.  All rights reserved.
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 */

#include "InterfaceInfoIterator.h"
#include "../Logger.h"
#include "../Error.h"
#include "Environment.h"


#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <linux/if.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ifaddrs.h>
#include <netdb.h>

static const char *INTERFACE_PREFIXES[] = { "enp", "ens", "eno", "eth", "tr", "wlan", "ppp", "bond", "vmnet", "em", "p1p", "p2p", "p3p", "p4p", "virbr", "ens3"};

InterfaceInfoIterator::InterfaceInfoIterator() : _ifaddrs(0), _ifaddrIt(0), _sd(-1), _prefixes(0)
{
    if (getifaddrs(&_ifaddrs) == -1) {
        _ifaddrs = 0;
        Error::setErrorMessage("Cannot obtain interfaces list");
        return;
    }
    _ifaddrIt = _ifaddrs;

    _sd = socket(PF_INET, SOCK_STREAM, 6);
    if (_sd < 0) {
        Error::setErrorMessage("Socket result less than 0");
        return;
    }
    initializeCustomInterfacePrefixes();
}

InterfaceInfoIterator::~InterfaceInfoIterator()
{
    if (_prefixes != 0) {
        freeCustomInterfacePrefixes();
    }
    if (_ifaddrs != 0) {
        freeifaddrs(_ifaddrs);
    }
    if (_sd >= 0) {
        close(_sd);
    }
}

void InterfaceInfoIterator::initializeCustomInterfacePrefixes()
{
    const char* envInterfaces = Environment::getenv("PARASOFT_SUPPORT_NET_INTERFACES");

    _numberOfAllPrefixes = (sizeof(INTERFACE_PREFIXES) / sizeof(char*));
    if (envInterfaces != 0) {
        const char* it;

        int length = 1;
        for (it = envInterfaces; *it; ++it) {
            if (*it == ',') {
                ++length;
            }
        }
        _numberOfCustomPrefixes = length;
        _numberOfAllPrefixes += _numberOfCustomPrefixes;
    } else {
        _numberOfCustomPrefixes = 0;
    }

    _prefixes = new char*[_numberOfAllPrefixes];
    int nextIdx = 0;

    if (envInterfaces != 0) {
        const char* it;
        const char* startInterfacePrefix = envInterfaces;
        for (it = envInterfaces; *it; ++it) {
            if (*it == ',') {
                _prefixes[nextIdx] = strndup(startInterfacePrefix, it - startInterfacePrefix);
                ++nextIdx;
                startInterfacePrefix = it + 1;
            }
        }
        _prefixes[nextIdx] = strndup(startInterfacePrefix, it - startInterfacePrefix);
        ++nextIdx;
    }
    for (unsigned i = 0; i < (sizeof(INTERFACE_PREFIXES) / sizeof(char*)); ++i) {
        _prefixes[nextIdx] = (char*)INTERFACE_PREFIXES[i];
        ++nextIdx;
    }
}

void InterfaceInfoIterator::freeCustomInterfacePrefixes()
{
    for (unsigned i = 0; i < _numberOfCustomPrefixes; ++i) {
        free(_prefixes[i]);
    }
    delete[] _prefixes;
}


bool InterfaceInfoIterator::hasNext()
{
    return _ifaddrIt != 0;
}

InterfaceInfo* InterfaceInfoIterator::next()
{
    InterfaceInfo* result = new InterfaceInfo();

    fillInterfaceInfo(result, _ifaddrIt);

    _ifaddrIt = _ifaddrIt->ifa_next;
    return result;
}

void InterfaceInfoIterator::fillInterfaceInfo(InterfaceInfo* result, ifaddrs *ifaddr)
{
    if (ifaddr->ifa_name != 0) {
        result->name = strdup(ifaddr->ifa_name);

        unsigned i = 0;
        for (; i < _numberOfAllPrefixes; ++i) {
            size_t prefixLength = strlen(_prefixes[i]);
            if (strncmp(result->name, _prefixes[i], prefixLength) == 0) {
                result->prefix = _prefixes[i];
                if (i < _numberOfCustomPrefixes) {
                    result->isCustom = true;
                }
                break;
            }
        }
        result->quality = i;
    } else {
        result->quality = _numberOfAllPrefixes;
    }

    if ((ifaddr->ifa_flags & IFF_LOOPBACK) == IFF_LOOPBACK) {
        result->isLoopBack = true;
    }

    bool isInet = (ifaddr->ifa_addr != 0) && (ifaddr->ifa_addr->sa_family == AF_INET);
    bool isInet6 = (ifaddr->ifa_addr != 0) && (ifaddr->ifa_addr->sa_family == AF_INET6);

    if (isInet) {
        result->lowQuality = 0;
    } else if (isInet6) {
        result->lowQuality = 1;
    } else {
        result->lowQuality = 2;
    }

    if ((ifaddr->ifa_name != 0) && (isInet || isInet6)) {
        struct ifreq ifr;
        memset(&ifr, 0, sizeof(struct ifreq));
        strncpy(ifr.ifr_name, ifaddr->ifa_name, sizeof(ifr.ifr_name));
        if (ioctl(_sd, SIOCGIFHWADDR, &ifr) == 0) {
            memcpy(result->mac, &ifr.ifr_hwaddr.sa_data, 6);
            for (int i = 0; i < 6; i++) {
                if (result->mac[i] != 0) {
                    result->hasMac = true;
                    break;
                }
            }
        }
        if (ioctl(_sd, SIOCGIFADDR, &ifr) == 0) {
            sockaddr_in* ipaddr = (sockaddr_in*)&ifr.ifr_addr;
            memcpy(result->ip, &ipaddr->sin_addr.s_addr, 4);
            for (int i = 0; i < 4; i++) {
                if (result->ip[i] != 0) {
                    result->hasIp = true;
                    break;
                }
            }
        }
    }
}
